#include <Xm/TextF.h>
#include <Xm/Form.h>
#include <Xm/ScrolledW.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/ToggleB.h>
#include <Xm/RowColumn.h>

#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <errno.h>
#include <malloc.h>
#include <string.h>

#include "timer.h"
#include "report.h"
#include "askuser.h"

#define MAGIC_STRING "PSP Event Log/1.1"
#define OLD_MAGIC  "PSP Event Log"
#define XMAX(a,b) ((a)>(b)?(a):(b))

#define LASTEVENT       (&(gEventList[gNumEvents - 1]))
#define EVENT_N(_n)     (&(gEventList[_n]))


Widget      gEventLogWindow;
Widget      gEventLogArea;
Widget      gEventLogScroll;


static tEvent          *gEventList;
static ulong            gNumEvents;
static char             gLastEventActive;

static char             gHaveInterruption;
static time_t           gInterruptionStart;

static char            *gCurrentLogFile;
char                    gEventLogIsDirty;
char                    gHaveEventLogFile;
ulong                   gMagic = 0xf85ec231;

void AddViewEntry(tEvent *);
void ClearViewEntries(void);
void UpdateViewEntries(void);
    
char            *gPhaseNames[] = { "P",
                                   "C",
                                   "B",
                                   "T",
                                   "M",
                                   "D",
                                   "d",
                                   "c"
                                   };
char            *gLongPhaseNames[] = { "Planning",
                                       "Coding",
                                       "Compiling",
                                       "Testing",
                                       "Postmortem",
                                       "Design",
                                       "Design review",
                                       "Code review"
                                       };


time_t Now(void)
{
    time_t  now;
    now = time(NULL);
    if (now == -1)
    {
        perror("time");
        now = 0;
    }
    return now;
}


void StartInterruption(void)
{
    if (gHaveInterruption == 0)
    {
        gHaveInterruption = 1;
        gInterruptionStart = Now();
    }
}

void EndInterruption(void)
{
    if (gHaveInterruption == 1)
    {
        if (gLastEventActive)
        {
            LASTEVENT->interruptionTime += Now() - gInterruptionStart;
        }
        gHaveInterruption = 0;
        gInterruptionStart = 0;
    }
}

void AllocEvent(void)
{
    void    *tmp;
    
    gNumEvents += 1;
    tmp = realloc(gEventList, gNumEvents * sizeof(*gEventList));
    if (tmp == NULL)
    {
        perror("realloc");
        exit(1);
    }
    gEventList = tmp;
}


void InitEvent(tEvent *event)
{
    event->start        = 0;
    event->end          = 0;
    event->interruptionTime = 0;
    event->eventType    = kNone;
    event->description  = NULL;
    event->exclude      = 0;
    event->index        = gNumEvents - 1;
}


void CloseEvent(void)
{
    if (gLastEventActive)
    {
        LASTEVENT->end = Now();
        if (gHaveInterruption)
        {
            LASTEVENT->interruptionTime +=
                LASTEVENT->end - gInterruptionStart;
            gInterruptionStart = Now();
        }
        AddViewEntry(LASTEVENT);
    }
    gLastEventActive = 0;
    gEventLogIsDirty = 1;
}

void AddEvent(void)
{
    CloseEvent();
    AllocEvent();
    InitEvent(LASTEVENT);
}

void NewEvent(tEventType eventType)

{
    AddEvent();
    LASTEVENT->start = Now();
    LASTEVENT->eventType = eventType;
    if (gHaveInterruption)
        gInterruptionStart = LASTEVENT->start;
    gLastEventActive = 1;
    gEventLogIsDirty = 1;
}


char *LookupPhase(tEventType phase)
{
    if (phase == kNone)
        return "-";
    else
        return gPhaseNames[(ulong)phase];
}

char *LongLookupPhase(tEventType phase)
{
    if (phase == kNone)
        return "--";
    else
        return gLongPhaseNames[(ulong)phase];
}


char ReadEventFile(char *fileName)
{
    FILE *file;
    char  buf[81], *s;
    int i, length;
    char haveReviewPhases;

    haveReviewPhases = 0;
    file = fopen(fileName, "rb");
    if (file == NULL)
    {
        Error("Can't open %s: %s", fileName, strerror(errno));
        return 0;
    }

    fgets(buf, sizeof(buf), file);
    buf[strlen(buf) - 1] = 0;
    
    if (strcmp(buf, MAGIC_STRING) && strcmp(buf, OLD_MAGIC))
    {
        Error("%s is not a PSP event log", fileName);
        fclose(file);
        return 0;
    }

    if (!strcmp(buf, OLD_MAGIC))
    {
        haveReviewPhases = 0;
    }
    else
    {
        haveReviewPhases = 1;
    }
        

    ResetEventLog();

    
    for (i = 0; i < NUM_TIMERS; i++)
    {
        gTimerList[i]->timerValue = 0;
        if (haveReviewPhases || (i != 2 && i != 4))
        {
            fscanf(file, "%lu ", &gTimerList[i]->timerValue);
        }
    }
    fscanf(file, "\n");


    fscanf(file, "%lu\n", &gNumEvents);
    gEventList = malloc(gNumEvents * sizeof(*gEventList));
    if (gEventList == NULL)
    {
        perror("alloc");
        exit(1);
    }

    for (i = 0; i < gNumEvents; i++)
    {
        fscanf(file, "%ld %ld %ld %d %d %dH",
               &(EVENT_N(i)->start),
               &(EVENT_N(i)->end),
               &(EVENT_N(i)->interruptionTime),
               (int *)&(EVENT_N(i)->eventType),
               &(EVENT_N(i)->exclude),
               &length);
        if (length)
        {
            s = malloc(length + 1);
            if (s == NULL)
            {
                perror("alloc");
                exit(1);
            }
            memset(s, 0, length + 1);
            if (fread(s, sizeof(*s), length, file) == length)
            {
                EVENT_N(i)->description = strdup(s);
            }
            else
            {
                Error("Failed to read description for an event");
            }
        }
        else
        {
            EVENT_N(i)->description = NULL;
        }
        fscanf(file, "\n");
    }
    
    fclose(file);

    UpdateViewEntries();
    
    return 1;
}

char WriteEventFile(char *fileName)
{
    FILE *file;
    int i;

    file = fopen(fileName, "wb");
    if (file == NULL)
    {
        Error("Can't open %s for writing: %s", fileName, strerror(errno));
        return 0;
    }

    /*
     * Print magic
     */
    
    fprintf(file, "%s\n", MAGIC_STRING);

    /*
     * Print timer values
     */
    
    for (i = 0; i < NUM_TIMERS; i++)
    {
        fprintf(file, "%lu ", gTimerList[i]->timerValue);
    }
    fprintf(file, "\n");

    /*
     * Print event log
     */

    fprintf(file, "%lu\n", gNumEvents);
    for (i = 0; i < gNumEvents; i++)
    {
        fprintf(file, "%ld %ld %ld %d %d %dH%s\n",
                EVENT_N(i)->start,
                EVENT_N(i)->end,
                EVENT_N(i)->interruptionTime,
                (int)(EVENT_N(i)->eventType),
                EVENT_N(i)->exclude,
                EVENT_N(i)->description ?
                strlen((EVENT_N(i))->description) : 0,
                EVENT_N(i)->description ?
                EVENT_N(i)->description : "");
    }

    if (ferror(file) != 0)
    {
        Error("Error while saving file: %s", strerror(ferror(file)));
        fclose(file);
        return 0;
    }
    fclose(file);

    return 1;
}

int SaveReport(char *fileName)
{
    FILE *file;

    file = fopen(fileName, "wb");
    if (file == NULL)
    {
        Error("Can't open %s for writing: %s", fileName, strerror(errno));
        return 0;
    }

    WriteEventLog(file);

    fclose(file);

    return 1;
}

void PrintReport(char *command)
{
    FILE *file;

    file = popen(command, "w");
    WriteEventLog(file);

    fclose(file);
}


void SaveTeXReport(char *fnam)
{
    FILE    *f, *file;
    char    *fileName, buf[1024];
    ulong       i;
    char        date[10];
    char        start[10];
    char        end[10];
    char        interrupt[10];
    char        total[10];
    struct tm   startTime, endTime, intrTime, totalTime;
    time_t      tempTime;

    file = fopen(fnam, "wb");
    if (file == NULL)
    {
        Error("Can't open %s for writing: %s", fileName, strerror(errno));
        return;
    }

    fileName = malloc(1 + strlen(PIXMAP_DIRECTORY) +
                      strlen("report.tex"));
    sprintf(fileName, "%s%s", PIXMAP_DIRECTORY, "report.tex");
    f = fopen(fileName, "rb");
    if (f == NULL)
    {
        Error("Unable to find LaTeX template for time\n"
              "reports. Your installation is probably\n"
              "broken. Have a nice day.");
        return;
    }
    while (!feof(f))
    {
        fgets(buf, sizeof(buf) - 1, f);
        if (!strncmp(buf, "% DATA", 6))
        {
            for (i = 0; i < gNumEvents; i++)
            {
                if (EVENT_N(i)->exclude)
                    continue;
                startTime = *localtime(&(EVENT_N(i)->start));
                endTime = *localtime(&(EVENT_N(i)->end));
                intrTime = *gmtime(&(EVENT_N(i)->interruptionTime));
                tempTime = EVENT_N(i)->end - EVENT_N(i)->start;
                totalTime = *gmtime(&tempTime);

                strftime(date, 7, "%y%m%d", &startTime);
                strftime(start, 6, "%H:%M", &startTime);
                strftime(end, 6, "%H:%M", &endTime);
                strftime(interrupt, 6, "%H:%M", &intrTime);
                strftime(total, 6, "%H:%M", &totalTime);

                fprintf(file, "%s & %s & %s & %s & %s & %s & %s\\\\\n",
                        date, start, end, interrupt, total,
                        LongLookupPhase(EVENT_N(i)->eventType),
                        EVENT_N(i)->description ?
                        EVENT_N(i)->description : "");
            }
        }
        else
            fprintf(file, "%s", buf);
    }
    fclose(f);
    fclose(file);

    return;
}

void WriteEventLog(FILE *file)
{
    ulong       i;
    char        date[10];
    char        start[10];
    char        end[10];
    char        interrupt[10];
    char        total[10];
    struct tm   startTime, endTime, intrTime, totalTime;
    time_t      tempTime;
    int         lineno, pageno;

    lineno = 0;
    pageno = 1;
    for (i = 0; i < gNumEvents; i++)
    {
        if (lineno == 0)
        {
            fprintf(file, "%s",
                    "   Date    Start    End     Intr   Total       Comment\n"
                    "+--------+-------+-------+-------+-------+---+"
                    "--------------------------------+\n");
            lineno += 2;
        }
        if (lineno == 62)
        {
            fprintf(file, 
                    "+--------+-------+-------+-------+-------+---+"
                    "--------------------------------+\n\n%79d", pageno);
            pageno += 1;
            lineno = 0;
            continue;
        }
        if (EVENT_N(i)->exclude)
            continue;
        
        startTime = *localtime(&(EVENT_N(i)->start));
        endTime = *localtime(&(EVENT_N(i)->end));
        intrTime = *gmtime(&(EVENT_N(i)->interruptionTime));
        tempTime = EVENT_N(i)->end - EVENT_N(i)->start;
        totalTime = *gmtime(&tempTime);

        strftime(date, 7, "%y%m%d", &startTime);
        strftime(start, 6, "%H:%M", &startTime);
        strftime(end, 6, "%H:%M", &endTime);
        strftime(interrupt, 6, "%H:%M", &intrTime);
        strftime(total, 6, "%H:%M", &totalTime);

        fprintf(file, "| %6s | %5s | %5s | %5s | %5s | %1s | %-30s |\n",
                date, start, end, interrupt, total,
                LookupPhase(EVENT_N(i)->eventType),
                EVENT_N(i)->description ? EVENT_N(i)->description : "");
        lineno += 1;
    }

    if (lineno != 0)
    {
        for (; lineno < 62; lineno++)
            fprintf(file, "| %6s | %5s | %5s | %5s | %5s | %1s | %-30s |\n",
                    "", "", "", "", "", "", "");
            fprintf(file, 
                    "+--------+-------+-------+-------+-------+---+"
                    "--------------------------------+\n\n%79d", pageno);
    }
}

void Testreport(void)
{
    WriteEventLog(stdout);
}


void ResetEventLog(void)
{
    free(gEventList);
    gEventList = NULL;
    gNumEvents = 0;
    gLastEventActive = 0;
    gHaveInterruption = 0;
    gInterruptionStart = 0;
    gEventLogIsDirty = 0;
    SetEventLogFilename (NULL);
    ClearViewEntries();
}


void SetEventLogFilename(char *fileName)
{
    if (gCurrentLogFile)
        free(gCurrentLogFile);
    
    if (fileName)
    {
        gCurrentLogFile = strdup(fileName);
        gHaveEventLogFile = 1;
    }
    else
    {
        gHaveEventLogFile = 0;
    }

    SetRightFooter(fileName ? fileName : "(None)");
}

char *GetEventLogFilename(void)
{
    return gCurrentLogFile;
}

void IncludeChangedCallback(Widget checkbox,
                            XtPointer clientData,
                            XtPointer callbackData)
{
    tEvent  *event;
    XmToggleButtonCallbackStruct    *cbs;

    event = EVENT_N((int)clientData);
    cbs = (XmToggleButtonCallbackStruct *)callbackData;

    event->exclude = !(cbs->set);

    gEventLogIsDirty = 1;
}

void CommentChangedCallback(Widget wodget,
                            XtPointer clientData,
                            XtPointer callbackData)
{
    tEvent      *event;
    char        *text;

    event = EVENT_N((int)clientData);

    XtVaGetValues(wodget,
                  XmNvalue, &text,
                  NULL);

    if (text)
    {
        if (event->description != NULL)
            free(event->description);
        event->description = strdup(text);
        XtFree(text);
        
        gEventLogIsDirty = 1;
    }
}

void ClearViewEntries(void)
{
    XtUnmanageChild(gEventLogArea);
    XtUnrealizeWidget(gEventLogArea);
    XtDestroyWidget(gEventLogArea);
    gEventLogArea = XtVaCreateManagedWidget("eventLogTable",
                                            xmRowColumnWidgetClass,
                                            gEventLogScroll,
                                            XmNisAligned, False,
                                            XmNorientation, XmVERTICAL,
                                            XmNnumColumns, 1,
                                            XmNpacking, XmPACK_COLUMN,
                                            XmNmarginHeight, 5,
                                            XmNmarginWidth, 5,
                                            NULL);
}

void UpdateViewEntries(void)
{
    int     i;

    ClearViewEntries();
    XtUnmanageChild(gEventLogArea);
    for (i = 0; i < gNumEvents; i++)
        AddViewEntry(EVENT_N(i));
    XtManageChild(gEventLogArea);
}


void AddViewEntry(tEvent *event)
{
    XmString        tmp;
    struct tm       startTime, endTime, intrTime, totalTime;
    time_t          tempTime;
    char            buf[80];
    char        date[10];
    char        start[10];
    char        end[10];
    char        interrupt[10];
    char        total[10];
    
    event->containerWidget = XtVaCreateWidget("eventLogEntry",
                                              xmFormWidgetClass,
                                              gEventLogArea,
                                              NULL);
    event->checkbox = XtVaCreateManagedWidget("includeField",
                                              xmToggleButtonWidgetClass,
                                              event->containerWidget,
                                              XmNindicatorType, XmN_OF_MANY,
                                              XmNindicatorOn, False,
                                              XmNselectPixmap,
                                              gCheckedPixmap,
                                              XmNlabelType, XmPIXMAP,
                                              XmNset, !event->exclude,
                                              XmNlabelPixmap,
                                              gUncheckedPixmap,
                                              XmNleftAttachment, XmATTACH_FORM,
                                              XmNtopAttachment, XmATTACH_FORM,
                                              XmNbottomAttachment,
                                              XmATTACH_FORM,
                                              NULL);

    XtAddCallback(event->checkbox,
                  XmNvalueChangedCallback,
                  IncludeChangedCallback,
                  (XtPointer)event->index);
                  
    
    startTime = *localtime(&(event->start));
    endTime = *localtime(&(event->end));
    intrTime = *gmtime(&(event->interruptionTime));
    tempTime = event->end - event->start;
    totalTime = *gmtime(&tempTime);
    
    strftime(date, 7, "%y%m%d", &startTime);
    strftime(start, 6, "%H:%M", &startTime);
    strftime(end, 6, "%H:%M", &endTime);
    strftime(interrupt, 6, "%H:%M", &intrTime);
    strftime(total, 6, "%H:%M", &totalTime);
    
    sprintf(buf, "%6s  %5s  %5s  %5s  %5s  %-10s",
            date, start, end, interrupt, total,
            LongLookupPhase(event->eventType));

    tmp = XmStringCreateSimple(buf);
    event->display = XtVaCreateManagedWidget("timeField",
                                             xmLabelWidgetClass,
                                             event->containerWidget,
                                             XmNlabelString, tmp,
                                             XmNleftAttachment,
                                             XmATTACH_WIDGET,
                                             XmNleftWidget, event->checkbox,
                                             XmNtopAttachment, XmATTACH_FORM,
                                             XmNbottomAttachment,
                                             XmATTACH_FORM,
                                             NULL);
    XmStringFree(tmp);

    event->comment = XtVaCreateManagedWidget("descriptionField",
                                             xmTextFieldWidgetClass,
                                             event->containerWidget,
                                             XmNcolumns,
                                             event->description?
                                             XMAX(strlen(event->description)
                                                 + 1, 30) : 30,
                                             XmNresizeWidth, True,
                                             XmNvalue,
                                             event->description ?
                                             event->description : "",
                                             XmNleftAttachment,
                                             XmATTACH_WIDGET,
                                             XmNleftWidget, event->display,
                                             XmNtopAttachment, XmATTACH_FORM,
                                             XmNbottomAttachment,
                                             XmATTACH_FORM,
                                             NULL);
    XtAddCallback(event->comment,
                  XmNvalueChangedCallback,
                  CommentChangedCallback,
                  (XtPointer)event->index);
                  
    XtManageChild(event->containerWidget);
}

void CreateEventLogWindow(Widget parent)
{
    Widget      form, header, containerWidget, checkbox, display, comment;
    XmString    tmp;
    char        buf[80];

    gEventLogWindow = XtVaCreatePopupShell("eventLogWindow",
                                           topLevelShellWidgetClass,
                                           parent,
                                           XmNdeleteResponse,
                                           XmUNMAP,
                                           NULL);
    form = XtVaCreateManagedWidget("mainArea",
                                   xmFormWidgetClass,
                                   gEventLogWindow,
                                   NULL);

    header = XtVaCreateManagedWidget("eventLogTableHeader",
                                     xmRowColumnWidgetClass,
                                     form,
                                     XmNisAligned, False,
                                     XmNnumColumns, 1,
                                     XmNpacking, XmPACK_COLUMN,
                                     XmNmarginWidth, 12,
                                     XmNtopAttachment, XmATTACH_FORM,
                                     XmNrightAttachment, XmATTACH_FORM,
                                     XmNleftAttachment, XmATTACH_FORM,
                                     NULL);
    
    containerWidget = XtVaCreateManagedWidget("eventLogEntry",
                                              xmFormWidgetClass,
                                              header,
                                              NULL);
    checkbox = XtVaCreateManagedWidget("includeField",
                                       xmLabelWidgetClass,
                                       containerWidget,
                                       XmNlabelType, XmPIXMAP,
                                       XmNlabelPixmap, gUncheckedPixmap,
                                       XmNleftAttachment, XmATTACH_FORM,
                                       XmNtopAttachment, XmATTACH_FORM,
                                       XmNbottomAttachment, XmATTACH_FORM,
                                       NULL);
    sprintf(buf, "%6s  %5s  %5s  %5s  %5s  %-10s",
            " Date ", "Start", " End ", "Inter", "Total", "Phase");

    tmp = XmStringCreateSimple(buf);
    display = XtVaCreateManagedWidget("timeField",
                                      xmLabelWidgetClass,
                                      containerWidget,
                                      XmNlabelString, tmp,
                                      XmNleftAttachment,
                                      XmATTACH_WIDGET,
                                      XmNleftWidget, checkbox,
                                      XmNtopAttachment, XmATTACH_FORM,
                                      XmNbottomAttachment,
                                      XmATTACH_FORM,
                                      NULL);
    XmStringFree(tmp);

    tmp = XmStringCreateSimple("Comment");
    comment = XtVaCreateManagedWidget("descriptionField",
                                      xmLabelWidgetClass,
                                      containerWidget,
                                      XmNleftAttachment,
                                      XmATTACH_WIDGET,
                                      XmNleftWidget, display,
                                      XmNtopAttachment, XmATTACH_FORM,
                                      XmNbottomAttachment,
                                      XmATTACH_FORM,
                                      XmNlabelString, tmp,
                                      NULL);
    XmStringFree(tmp);


    gEventLogScroll = XtVaCreateManagedWidget("eventLogScrolledWindow",
                                              xmScrolledWindowWidgetClass,
                                              form,
                                              XmNtopAttachment,
                                              XmATTACH_WIDGET,
                                              XmNtopWidget,
                                              header,
                                              XmNrightAttachment,
                                              XmATTACH_FORM,
                                              XmNleftAttachment, XmATTACH_FORM,
                                              XmNbottomAttachment,
                                              XmATTACH_FORM,
                                              XmNscrollBarDisplayPolicy,
                                              XmAS_NEEDED,
                                              XmNscrollingPolicy,
                                              XmAUTOMATIC,
                                              NULL);
    gEventLogArea = XtVaCreateManagedWidget("eventLogTable",
                                            xmRowColumnWidgetClass,
                                            gEventLogScroll,
                                            XmNisAligned, False,
                                            XmNorientation, XmVERTICAL,
                                            XmNnumColumns, 1,
                                            XmNpacking, XmPACK_COLUMN,
                                            NULL);

}

void ViewEventLogWindow(void)
{
    XtPopup(gEventLogWindow,
            XtGrabNone);
}
